package myhr.facescan.myhr_facescan

import android.graphics.Bitmap
import android.graphics.BitmapFactory
import android.graphics.Matrix
import androidx.annotation.NonNull

import io.flutter.embedding.engine.plugins.FlutterPlugin
import io.flutter.plugin.common.MethodCall
import io.flutter.plugin.common.MethodChannel
import io.flutter.plugin.common.MethodChannel.MethodCallHandler
import io.flutter.plugin.common.MethodChannel.Result
import java.io.ByteArrayOutputStream
import kotlinx.coroutines.GlobalScope
import kotlinx.coroutines.launch

/** MyhrFacescanPlugin */
class MyhrFacescanPlugin : FlutterPlugin, MethodCallHandler {
    /// The MethodChannel that will the communication between Flutter and native Android
    ///
    /// This local reference serves to register the plugin with the Flutter Engine and unregister it
    /// when the Flutter Engine is detached from the Activity
    private lateinit var channel: MethodChannel

    override fun onAttachedToEngine(flutterPluginBinding: FlutterPlugin.FlutterPluginBinding) {
        channel = MethodChannel(flutterPluginBinding.binaryMessenger, "myhr_facescan")
        channel.setMethodCallHandler(this)
    }

    override fun onMethodCall(call: MethodCall, result: Result) {
        if (call.method == "activateLuxand") {
            result.success(activateLuxand(call.arguments))
        } else if (call.method == "detectFace") {
            GlobalScope.launch {
                val found = detectFace(call.arguments)
                result.success(found)
            }
        } else if (call.method == "getFaceTemplate") {
            val templates = getFaceTemplate(call.arguments)
            var s1 = templates.get(0)?.decodeToString()
            var s2 = templates.get(1)?.decodeToString()
            var stringArrays = arrayListOf(s1, s2)
            result.success(stringArrays)
        } else if (call.method == "getFaceAttribute") {
            GlobalScope.launch {
                val data = getFaceAttribute(call.arguments)
                result.success(data)
            }
        } else if (call.method == "freeResource") {
            LuxandSdk.FreeTracker()
        } else {
            result.notImplemented()
        }
    }

    private fun activateLuxand(data: Any): Int {
        val key = data as Map<String, String>
        val licenseKey = key["licenseKey"] as String
        val libCode = LuxandSdk.Initialize(licenseKey);

        return libCode;
    }

    private fun getImageBytes(data: Any): ByteArray {
        val key = data as Map<String, String>
        val bytesList = key["platforms"] as List<ByteArray>;
        val strides = key["strides"] as IntArray;
        val width = key["width"] as Int;
        val height = key["height"] as Int;
        val imageBytes = YuvConverter.NV21toJPEG(
            YuvConverter.YUVtoNV21(
                bytesList, strides, width, height
            ), width, height, 80
        )

        return imageBytes
    }

    private fun detectFace(data: Any): IntArray {
        val imageBytes = getImageBytes(data)
        val rotateBitmap = rotateImage(byteArrayToBitmap(imageBytes), 270f)
        val bytes = bitmapToByteArray(rotateBitmap, Bitmap.CompressFormat.JPEG, 100)

        return LuxandSdk.TrackFace(bytes)
    }

    private fun getFaceTemplate(data: Any): Array<ByteArray> {
        val imageBytes = getImageBytes(data)
        val rotateBitmap = rotateImage(byteArrayToBitmap(imageBytes), 270f)
        val result = bitmapToByteArray(rotateBitmap, Bitmap.CompressFormat.JPEG, 100)
        val templates = LuxandSdk.GetTemplate(result)

        return templates
    }

    private fun getFaceAttribute(data: Any): DoubleArray {
        val imageBytes = getImageBytes(data)
        val rotateBitmap = rotateImage(byteArrayToBitmap(imageBytes), 270f)
        val result = bitmapToByteArray(rotateBitmap, Bitmap.CompressFormat.JPEG, 100)
        val key = data as Map<String, String>
        val isExpression = key["isExpression"] as Boolean;
        return LuxandSdk.GetFaceAttribute(result, isExpression)
    }

    fun bitmapToByteArray(bitmap: Bitmap, format: Bitmap.CompressFormat, quality: Int): ByteArray? {
        val stream = ByteArrayOutputStream()
        if (!bitmap.compress(format, quality, stream)) {
            return null
        }
        return stream.toByteArray()
    }

    fun byteArrayToBitmap(byteArray: ByteArray): Bitmap {
        return BitmapFactory.decodeByteArray(byteArray, 0, byteArray.size)
    }

    fun rotateImage(bitmap: Bitmap, degrees: Float): Bitmap {
        val matrix = Matrix()
        matrix.postRotate(degrees)
        matrix.postScale(-1f, 1f, bitmap.width / 2f, bitmap.height / 2f);

        val rotatedBitmap = Bitmap.createBitmap(
            bitmap, 0, 0, bitmap.width, bitmap.height, matrix, true
        )

        return rotatedBitmap
    }

    override fun onDetachedFromEngine(binding: FlutterPlugin.FlutterPluginBinding) {
        channel.setMethodCallHandler(null)
    }
}
