import 'dart:async';
import 'dart:io';
import 'dart:math';

import 'package:camera/camera.dart';
import 'package:flutter/material.dart';
import 'package:get/get.dart';
import 'package:image/image.dart' as imglib;
import 'package:myhr_facescan/enroll/template_model.dart';
import 'package:myhr_facescan/luxand_controller.dart';
import 'package:myhr_facescan/throttler.dart';

class EnrollController extends GetxController {
  EnrollController({required this.licenseKey});

  CameraController? camController;
  final _throttler = Throttler(milliSeconds: 100);
  // StreamSubscription<int> _timer = null;

  final String licenseKey;

  var isFoundFace = false.obs;
  var faceX = 0.0.obs;
  var faceY = 0.0.obs;
  var faceW = 0.0.obs;
  var isCameraReady = false.obs;
  var faces = <Image>[].obs;
  var age = 0.obs;
  var isMale = true.obs;
  var currentCommand = 0.obs;
  var livenessMode = false.obs;

  var commands = <ECommand>[];
  var templates = <TemplateModel>[];
  var gettingCurrentImage = false;
  LuxandController? luxand;

  get commandName {
    switch (commands[currentCommand.value]) {
      case ECommand.lookStraight:
        return 'มองตรง';
      case ECommand.turnLeft:
        return 'หันไปทางซ้าย';
      case ECommand.turnRight:
        return 'หันไปทางขวา';
      case ECommand.lookUp:
        return 'เงยหน้า';
      case ECommand.lookDown:
        return 'ก้มหน้า';
      case ECommand.smile:
        return 'ยิ้ม';
      default:
        return 'N/A';
    }
  }

  approveCommand() async {
    if (currentCommand.value + 1 >= commands.length) {
      currentCommand.value = 0;
    } else {
      currentCommand.value += 1;
    }

    camController!.stopImageStream();
    await Future.delayed(const Duration(seconds: 3));
    camController!.startImageStream(onAvailable);
  }

  initCamera() async {
    final cameraDescription = (await availableCameras())
        .where(
          (element) => element.lensDirection == CameraLensDirection.front,
        )
        .first;

    camController = CameraController(cameraDescription, ResolutionPreset.high,
        enableAudio: false,
        imageFormatGroup: Platform.isIOS
            ? ImageFormatGroup.bgra8888
            : ImageFormatGroup.yuv420);
  }

  onAvailable(CameraImage image) async {
    _throttler.run(() async {
      if (livenessMode.value) {
        var cm = commands[currentCommand.value];
        var isExpression = cm == ECommand.smile;
        var attr = await luxand!.getFaceAttribute(image, isExpression);

        if (isExpression) {
          // const double SMILE_CONF_LEVEL_LOW = 0.3;
          const double SMILE_CONF_LEVEL_HIGH = 0.6;

          if (attr[0] > SMILE_CONF_LEVEL_HIGH && cm == ECommand.smile) {
            print(
                '========================================== Smile attr=${attr[0]}, SMILE_CONF_LEVEL_HIGH=$SMILE_CONF_LEVEL_HIGH');
            await approveCommand();
          }
        } else {
          if (cm == ECommand.turnLeft ||
              cm == ECommand.turnRight ||
              cm == ECommand.lookStraight) {
            const double HOR_CONF_LEVEL_LOW = 2;
            const double HOR_CONF_LEVEL_HIGH = 17;

            if (attr[0].abs() > HOR_CONF_LEVEL_HIGH) {
              if (attr[0] > 0 && cm == ECommand.turnLeft) {
                print(
                    '========================================== Turn Left attr=${attr[0]}, HOR_CONF_LEVEL_HIGH=$HOR_CONF_LEVEL_HIGH');
                await approveCommand();
              } else if (attr[0] <= 0 && cm == ECommand.turnRight) {
                print(
                    '========================================== Turn Right attr=${attr[0]}, HOR_CONF_LEVEL_HIGH=$HOR_CONF_LEVEL_HIGH');
                await approveCommand();
              }
            } else if (attr[0].abs() < HOR_CONF_LEVEL_LOW &&
                cm == ECommand.lookStraight) {
              print(
                  '========================================== Look Straight attr=${attr[0]}, HOR_CONF_LEVEL_LOW=$HOR_CONF_LEVEL_LOW');
              await approveCommand();
            }
          } else {
            // const double UP_CONF_LEVEL_LOW = 2;
            const double UP_CONF_LEVEL_HIGH = 6;
            // const double DOWN_CONF_LEVEL_LOW = -2;
            const double DOWN_CONF_LEVEL_HIGH = -3;

            if (attr[1] > UP_CONF_LEVEL_HIGH && cm == ECommand.lookUp) {
              print(
                  '========================================== Look Up attr=${attr[1]}, UP_CONF_LEVEL_HIGH=$UP_CONF_LEVEL_HIGH');
              await approveCommand();
            } else if (attr[1] < DOWN_CONF_LEVEL_HIGH &&
                cm == ECommand.lookDown) {
              print(
                  '========================================== Look Down attr=${attr[1]}, DOWN_CONF_LEVEL_HIGH=$DOWN_CONF_LEVEL_HIGH');
              await approveCommand();
            }
          }
        }
      } else {
        if (gettingCurrentImage) {
          if (!isFoundFace.value) {
            return;
          }

          var ft = await luxand!.getFaceTemplate(image);
          var png = await convertYUV420toImageColor(image);

          if (png != null) {
            templates.add(TemplateModel(face: ft[1], eye: ft[0]));

            faces.value.add(png);
            faces.refresh();

            if (faces.length >= 3) {
              livenessMode.value = true;
              isFoundFace.value = false;

              commands = getCommands();
            }
          }

          gettingCurrentImage = false;
        } else {
          var data = await luxand!.detectFace(image);

          if (data == null) {
            return;
          }

          var isFound = data[0] > 0;
          isFoundFace.value = isFound;

          if (isFound) {
            faceX.value = data[0].toDouble();
            faceY.value = data[1].toDouble();
            faceW.value = data[2].toDouble();
            age.value = data[3];
            isMale.value = data[4] == 1;
          }
        }
      }
    });
  }

  List<ECommand> getCommands() {
    var result = [ECommand.lookStraight];

    while (true) {
      if (result.contains(ECommand.lookUp) &&
          result.contains(ECommand.lookDown) &&
          result.contains(ECommand.turnLeft) &&
          result.contains(ECommand.turnRight) &&
          result.contains(ECommand.smile)) {
        break;
      }

      do {
        var cm = ECommand.fromInt(Random().nextInt(6));

        if (!result.contains(cm) && cm != ECommand.lookStraight) {
          result.add(cm);
          result.add(ECommand.lookStraight);
          break;
        }
      } while (true);
    }

    result.removeLast();

    return result;
  }

  @override
  void onInit() async {
    super.onInit();

    luxand = Get.put(LuxandController(licenseKey: licenseKey));

    await luxand!.initialize();
    await initCamera();

    camController!.initialize().then((_) {
      isCameraReady.value = true;

      // Only open and close camera in iOS for low-tier device
      if (Platform.isIOS) {
        // _timer = Stream.periodic(const Duration(milliseconds: 500), (v) => v)
        //     .listen((count) async {
        //   _throttler.run(() async {
        //     controller.startImageStream((image) async {});

        //     Future.delayed(const Duration(milliseconds: 50), () async {
        //       await controller.stopImageStream();
        //     });
        //   });
        // });
      } else {
        camController!.startImageStream(onAvailable);
      }
    });
  }

  getCurrentImage() async {
    gettingCurrentImage = true;
  }

  Future<Image?> convertYUV420toImageColor(CameraImage image) async {
    try {
      final int width = image.width;
      final int height = image.height;
      final int uvRowStride = image.planes[1].bytesPerRow;
      final int uvPixelStride = image.planes[1].bytesPerPixel!;
      var img = imglib.Image(height: width, width: height);

      // Fill image buffer with plane[0] from YUV420_888
      for (int x = 0; x < width; x++) {
        for (int y = 0; y < height; y++) {
          final int uvIndex =
              uvPixelStride * (x / 2).floor() + uvRowStride * (y / 2).floor();
          final int index = y * width + x;

          final yp = image.planes[0].bytes[index];
          final up = image.planes[1].bytes[uvIndex];
          final vp = image.planes[2].bytes[uvIndex];
          // Calculate pixel color
          int r = (yp + vp * 1436 / 1024 - 179).round().clamp(0, 255);
          int g = (yp - up * 46549 / 131072 + 44 - vp * 93604 / 131072 + 91)
              .round()
              .clamp(0, 255);
          int b = (yp + up * 1814 / 1024 - 227).round().clamp(0, 255);
          // color: 0x FF  FF  FF  FF
          //           A   B   G   R
          // img.data[index] = shift | (b << 16) | (g << 8) | r;

          if (img.isBoundsSafe(height - y, x)) {
            const shift = (0xFF << 24);
            img.setPixelRgba(height - y, x, r, g, b, shift);
          }
        }
      }

      img = imglib.copyRotate(img, angle: -180);
      img = imglib.flipHorizontal(img);

      var encoder = imglib.JpegEncoder();
      var bytes = encoder.encode(img);

      return Image.memory(bytes);
    } catch (e) {
      print(">>>>>>>>>>>> ERROR:$e");
    }

    return null;
  }

  @override
  void onClose() async {
    super.onClose();

    camController?.stopImageStream();
    camController?.dispose();

    await luxand!.freeResource();
  }
}

enum ECommand {
  lookStraight(0),
  lookUp(1),
  lookDown(2),
  turnLeft(3),
  turnRight(4),
  smile(5);

  final int value;

  const ECommand(this.value);

  factory ECommand.fromInt(int value) {
    return values.firstWhere((light) => light.value == value,
        orElse: () => ECommand.lookStraight);
  }
}
